/*
 *  aatv : a simple program to view TV on a text console
 *  Copyright (C) 2000  Florent de LAMOTTE <florent@lamotte.as>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Picked some code and information from 
 *  XawTV (c) 1998-2000 Gerd Knorr 
 *  you can download it at http://www.strusel007.de/linux/xawtv
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <aalib.h>

#include <asm/types.h>
#include <linux/videodev.h>
 
char *grab_device = "/dev/video0";
int   grab_width  = 640;
int   grab_height = 480;
int   grab_input = 0;
int   grab_norm  = VIDEO_MODE_SECAM;
unsigned long presets [10] = {0,679250,655250,631250,0,703250,727250,0,0,0};


static struct video_capability   grab_cap;
static struct video_mmap         grab_buf;
static struct video_channel	 grab_chan;
static struct video_picture      grab_pict;
static struct video_window       grab_win;
static struct video_audio        grab_audio;
static int                       grab_fd, grab_size, have_mmap;
static int                       mix;
static unsigned char            *grab_data;

void
parse_config ()
{
     char line [256];     
     char filename [100];
     char *buf;
     int chan;
     FILE *fp;
     
     strcpy (filename, getenv ("HOME"));
     strcat (filename, "/.aatv");

     if ((fp = fopen(filename, "r")) == NULL) 
	  return;


     /* parse file */
     while ((fgets(line, 255, fp)) != NULL) {
	  buf = strtok (line, " ");
	  if (buf[0] == '\n' || buf[0] == '#') 
	       continue; /* empty or commented */
	  if (!strcasecmp(buf, "channel")) {
	       chan = atoi(strtok(NULL," "));
	       if (chan >= 0 && chan <= 9)
		    presets [chan] = (unsigned long) atol(strtok(NULL," "));
	  } else if (!strcasecmp (buf, "vidmode")) {
	       buf = strtok (NULL, " ");
	       if (!strcasecmp (buf, "PAL"))
		    grab_norm = VIDEO_MODE_PAL;
	       else if (!strcasecmp (buf, "SECAM"))
		    grab_norm = VIDEO_MODE_SECAM;
	       else if (!strcasecmp (buf, "NTSC"))
		    grab_norm = VIDEO_MODE_NTSC;
	  }
     }	       
     fclose (fp);
}

void
grab_init ()
{
     int devmask;

     if ((grab_fd = open(grab_device,O_RDWR)) == -1) {
	  fprintf(stderr,"open %s: %s\n",grab_device,strerror(errno));
	  exit(1);
     }
     if (ioctl(grab_fd,VIDIOCGCAP,&grab_cap) == -1) {
	  fprintf(stderr,"%s: no v4l device\n",grab_device);
	  exit(1);
     }

     /* set image source and TV norm */
     grab_chan.channel = grab_input;
     if (ioctl(grab_fd,VIDIOCGCHAN,&grab_chan)==-1) {
	  perror("ioctl VIDIOCGCHAN");
	  exit(1);
     }
     grab_chan.channel = grab_input;
     grab_chan.norm    = grab_norm;
     if (ioctl(grab_fd,VIDIOCSCHAN,&grab_chan)==-1) {
	  perror("ioctl VIDIOCSCHAN");
	  exit(1);
     }

     /* open the mixer */
     if ((mix = open("/dev/mixer", O_RDONLY)) == -1) {
	  perror("mixer open");
     }
     fcntl(mix,F_SETFD,FD_CLOEXEC);

     if (-1 == ioctl(mix,MIXER_READ(SOUND_MIXER_DEVMASK),&devmask)) {
	  perror("mixer read devmask");
     }
     
     close(mix);

     /* turn audio on */
     if (ioctl(grab_fd,VIDIOCGAUDIO,&grab_audio) == -1) {
	  perror("ioctl VIDIOCGAUDIO");
	  exit(1);
     }
    
    
     grab_audio.flags &= ~VIDEO_AUDIO_MUTE;

     if (ioctl(grab_fd,VIDIOCSAUDIO,&grab_audio) == -1) {
	  perror("ioctl VIDIOCSAUDIO");
	  exit(1);
     }


     /* try to setup mmap-based capture */
     grab_buf.format = VIDEO_PALETTE_GREY;
     grab_buf.frame  = 0;
     grab_buf.width  = grab_width;
     grab_buf.height = grab_height;
     grab_size = grab_buf.width * grab_buf.height;
     grab_data = mmap(0,grab_size,PROT_READ|PROT_WRITE,MAP_SHARED,grab_fd,0);
     if ((int)grab_data != -1) {
	  have_mmap = 1;
	  return;
     }

     /* fallback to read() */
     fprintf(stderr,"no mmap support available, using read()\n");
     have_mmap = 0;
     grab_pict.depth   = 8;
     grab_pict.palette = VIDEO_PALETTE_GREY;
     if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) == -1) {
	  perror("ioctl VIDIOCSPICT");
	  exit(1);
     }
     if (ioctl(grab_fd,VIDIOCGPICT,&grab_pict) == -1) {
	  perror("ioctl VIDIOCGPICT");
	  exit(1);
     }
     memset(&grab_win,0,sizeof(struct video_window));
     grab_win.width  = grab_width;
     grab_win.height = grab_height;
     if (ioctl(grab_fd,VIDIOCSWIN,&grab_win) == -1) {
	  perror("ioctl VIDIOCSWIN");
	  exit(1);
     }
     if (ioctl(grab_fd,VIDIOCGWIN,&grab_win) == -1) {
	  perror("ioctl VIDIOCGWIN");
	  exit(1);
     }

     grab_size = grab_win.width * grab_win.height;
     grab_data = malloc(grab_size);
}

unsigned char*
grab_one(int *width, int *height)
{
     int rc;
    
     for (;;) {
	  if (have_mmap) {
	       if (-1 == ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buf)) {
		    perror("ioctl VIDIOCMCAPTURE");
	       } else {
		    if (-1 == ioctl(grab_fd,VIDIOCSYNC,&grab_buf)) {
			 perror("ioctl VIDIOCSYNC");
		    } else {
			 *width  = grab_buf.width;
			 *height = grab_buf.height;
			 return grab_data;
		    }
	       }
	  } else {
	       rc = read(grab_fd,grab_data,grab_size);
	       if (grab_size != rc) {
		    fprintf(stderr,"grabber read error (rc=%d)\n",rc);
		    return NULL;
	       } else {
		    *width  = grab_win.width;
		    *height = grab_win.height;
		    return grab_data;
	       }
	  }
     }
}


/* fo the window */
aa_context *c;
aa_renderparams *p;

int
(*show_image) (char *data, int width, int height);

int 
show_image_std (char *data, int width, int height)
{
     int i=0;
     int j=height*width/4;
     while (i<j) ((unsigned long*)c->imagebuffer)[i] =((unsigned long*)data)[i++];
}

int
show_image_invert (char *data, int width, int height)
{
     int i,j,k;
     for (i=width-1,k=0;i>=0;i--,k++) 
	  for (j=0;j<height;j++)
	       aa_putpixel (c,i,j,data[k+j*width]);
}
     
unsigned long
get_freq ()
{
     unsigned long freq;
     if (ioctl(grab_fd, VIDIOCGFREQ, &freq) == -1)
	  perror("ioctl VIDIOCGFREQ");
     return (freq * 125 / 2);
}
     
int
change_freq (unsigned long freq)
{
     freq = freq * 2 / 125;
     if (ioctl(grab_fd, VIDIOCSFREQ, &freq) == -1)
	  perror("ioctl VIDIOCSFREQ");
     return 0;
}

void
set_picture_properties ()
{
     if (ioctl(grab_fd, VIDIOCSPICT, &grab_pict) == -1)
	  perror ("ioctl VIDIOSPICT");
     if (ioctl(grab_fd, VIDIOCGPICT, &grab_pict) == -1)
	  perror ("ioctl VIDIOSPICT");
}

void
get_picture_properties ()
{
     if (ioctl(grab_fd, VIDIOCGPICT, &grab_pict) == -1)
	  perror ("ioctl VIDIOSPICT");
}
     

/* ---------------------------------------------------------------------- */


int
main(int argc, char *argv[])
{
     unsigned char *image;
     int width, height, i;
     int inverse;
     int cont = 1;		/* flag to end main loop */
     int key;
     int fast = 0;		/* fast render ON/OFF */
     char buf [10];		/* buffer for interactive commands */
    

     parse_config ();

     /* initialize AA-lib */
     if (!aa_parseoptions(NULL, NULL, &argc, argv) || argc != 1) {
	  printf("%s\n", aa_help);
	  exit(1);
     }

     c = aa_autoinit(&aa_defparams);
     if (c == NULL) {
	  printf("Can not intialize aalib\n");
	  exit(2);
     }
     if (!aa_autoinitkbd(c,0)) {
	  printf("Can not intialize keyboard\n");
	  aa_close(c);
	  exit(3);
     }
     aa_hidemouse (c);

     /* init grabber */
     p = aa_getrenderparams();
     grab_width = aa_imgwidth (c);
     grab_height = aa_imgheight (c);
     show_image = &show_image_std;
     inverse=0;
     
     grab_init();
     get_picture_properties ();

     /* main loop */
     while (cont) {
	  image = grab_one(&width,&height);
	  (*show_image)(image,width,height);
	  if (fast)
	       aa_fastrender(c, 0, 0, aa_scrwidth(c), aa_scrheight(c));
	  else
	       aa_render(c, p, 0, 0, aa_scrwidth(c), aa_scrheight(c));
	  aa_flush(c);

	  if ((key=aa_getkey(c,0)) == AA_NONE)
	       continue;

	  switch (key) {
	  case 'q':
	       cont = 0;
	       break;
	  case 'v': 
	       if (inverse=!inverse)
		    show_image=&show_image_invert;
	       else
		    show_image=&show_image_std;
	       break;
	  case 'c':
	       if (!grab_chan.channel) {
		    grab_chan.channel = 1; 
		    grab_chan.norm    = VIDEO_MODE_PAL;
		    if (-1 == ioctl(grab_fd,VIDIOCSCHAN,&grab_chan)) {
			 perror("ioctl VIDIOCSCHAN");
			 exit(1);
		    }
	       } else { /* switch from camera to TV */
		    grab_chan.channel = 0;
		    grab_chan.norm = grab_norm;
		    if (-1 == ioctl(grab_fd,VIDIOCSCHAN,&grab_chan)) {
			 perror("ioctl VIDIOCSCHAN");
			 exit(1);
		    }
	       }
	       break;

	  case 's' :
	       if (++grab_norm == 3)
		    grab_norm = 0;
	       grab_chan.norm = grab_norm;
	       if (-1 == ioctl(grab_fd,VIDIOCSCHAN,&grab_chan)) {
		    perror("ioctl VIDIOCSCHAN");
		    exit(1);
	       }
	       break;
	       
	       /* AA image controls */
	  case 'f':
	       fast=!fast;
	       break;
	  case 'a':
	       p->dither=AA_NONE;
	       break;
	  case 'z':
	       p->dither=1;
	       break;
	  case 'e':
	       p->dither=2;
	       break;
	  case 'p':		/* contrast */
	       sprintf(buf,"%d",p->contrast);
	       aa_edit(c,0,0,5,buf,5);
	       p->contrast=atol(buf);
	       break; 
	  case '!':		/* contrast */
	       p->contrast--;
	       break;
	  case 'm':		/* contrast */
	       p->contrast++;
	       break;		/* brightness */
	  case 'o':
	       sprintf(buf,"%d",p->bright);
	       aa_edit(c,0,0,5,buf,5);
	       p->bright=atol(buf);
	       break;
	  case 'l':		/* brightness */
	       p->bright++;
	       break;
	  case ':':		/* brightness */
	       p->bright--;
	       break;

	       /* Tuner Image control */
	  case 'i':		/* contrast */
	       sprintf(buf,"%d",grab_pict.contrast);
	       aa_edit(c,0,0,6,buf,6);
	       grab_pict.contrast=atoi(buf);
	       set_picture_properties ();
	       break;
	  case 'k':		/* contrast */
	       if (grab_pict.contrast<65535)
		    grab_pict.contrast++;
	       set_picture_properties ();
	       break;
	  case ';':		/* contrast */
	       if (grab_pict.contrast>0)
		    grab_pict.contrast--;
	       set_picture_properties ();
	       break;
	  case 'u':		/* brightness */
	       sprintf(buf,"%d",grab_pict.brightness);
	       aa_edit(c,0,0,6,buf,6);
	       grab_pict.brightness=atoi(buf);
	       set_picture_properties ();
	       break;
	  case 'j':		/* brightness */
	       if (grab_pict.brightness<65535)
		    grab_pict.brightness++;
	       set_picture_properties ();
	       break;
	  case ',':		/* brightness */
	       if (grab_pict.brightness>0)
		    grab_pict.brightness--;
	       set_picture_properties ();
	       break;
	  case 'y':		/* whiteness */
	       sprintf(buf,"%d",grab_pict.whiteness);
	       aa_edit(c,0,0,6,buf,6);
	       grab_pict.whiteness=atoi(buf);
	       set_picture_properties ();
	       break;
	  case 'h':		/* whiteness */
	       if (grab_pict.whiteness<65535)
		    grab_pict.whiteness++;
	       set_picture_properties ();
	       break;
	  case 'n':		/* whiteness */
	       if (grab_pict.whiteness>0)
		    grab_pict.whiteness--;
	       set_picture_properties ();
	       break;

	       /* frequencies */
	  case 'b':
	       sprintf(buf,"%d",get_freq());
	       aa_edit (c, 0,0,7,buf,7);
	       change_freq (atol(buf));
	       break;
	  default :
	       if ((key >= '0') && (key <= '9') && (presets [key-'0']!=0)) {
		    change_freq (presets [key - '0']);
	       }
	       break;
	  }
     }

     aa_uninitmouse(c);
     aa_close (c);

     /* turn off audio */
     grab_audio.flags |= VIDEO_AUDIO_MUTE;
     if (ioctl(grab_fd,VIDIOCSAUDIO,&grab_audio) == -1) {
	  perror("ioctl VIDIOCSAUDIO");
	  exit(1);
     }

     return 0;
}

